/*  
  Edit this for CTCSS frequencies wanted. NOTE: Converted to int to save space
  Frequ synth for Phillips FM747U
  Denys VK3ZYZ (with help from Google and others)
  Fixes by Ray and Josh 20190728
  CTCSS tones added :)
  displays CTCSS tone frequ at start of top line when on TX 20191007.
  Modified to suite this radio by Ian VK3YYY 2020 01 20
  Memory fix - subtone to int 2020 02 18
  */

#define Version "Phillips FM 747U 20200903"

#include "si5351.h"
#include "Wire.h"
#include "LiquidCrystal_PCF8574.h"

// synthesiser default settings on start up.
unsigned long RXgenfrequency  = 433025000;
unsigned long TXgenfrequency  = 433025000;
unsigned long IFoffset        = 29800000;    // IF offset
unsigned long TXmultiplier    = 24;          // transmitter frequency multiplier
unsigned long RXmultiplier    = 9;           // receiver frequency multiplier

float         txDisplayFreq   = 433.025;
float         rxDisplayFreq   = 433.025;
int           mode            = HIGH;        // HIGH = normal, LOW = REV
int           TRXmode         = HIGH;        // HIGH = RX, LOW = TX
int           channel         = 0;           // convert frequence to channel number
int encoder;                                 // encoder direction: CCW=-1; NC=0; CW=1
int timer = 100;           // sub tone setting. The higher the number, the slower the timing.
int  i ;

Si5351 clockgen;                             // initalize clockgen

// encoder definitions
#define ENCODER_A     A0
#define ENCODER_B     A1
//#define SSB_SWITCH  A2           // HIGH = LSB, LOW = USB.
//#define AM_SWITCH   A3           // not used yet
// A4 = SDA
// A5 = SCL
#define AnalogIn1     A6           // not used yet
#define AnalogIn2     A7           // not used yet
#define spare         A2           // not used yet
#define TRX            2           // input HIGH = RX, LOW = TX.
#define Input2         3           // not used yet
#define Output1        4           // not used yet
#define REV_SWITCH     5           // Encoder Button  repeater reverse
#define spareSW        6           // not used yet

/* ********************* PortB0-PB5 used to generate CTCSS tones via R2R nerwork.
#define spare1         7           // Sub Tone R2R b0 PB0
#define spare2         8           // Sub Tone R2R b1 PB1
#define spare3         9           // Sub Tone R2R b2 PB2
#define spare4        10           // Sub Tone R2R b3 PB3
#define spare5        11           // Sub Tone R2R b4 PB4
#define spare6        12           // Sub Tone R2R b5 PB5
*/

#define Output2       13           // not used yet

#define ENCODER_INPUT_PORT PINC
#define On  1
#define Off 0
#define RefFrequ      500000        // 16Mhz / 32 steps for CTCSS sine wave generator
int     SineDivider = 4065;         // default sine wave divider
#define MaxChannel 72              // Tables must be the same length!!!!

 
long int RXFrequs []      // Edit this for RX frequencies wanted.
{ 146000,
  146625,
  146650,
  146675,
  146700,
  146725,
  146750,
  146775,
  146800,
  146825,
  146850,
  146875,
  146900,
  146925,
  146950,
  146975,
  147000,
  146425,
  146450,
  146475,
  146500,
  146525,
  146550,
  146575,
  146600,
  147025,
  147050,
  147075,
  147100,
  147125,
  147150,
  147175,
  147200,
  147225,
  147250,
  147275,
  147300,
  147325,
  147350,
  147375,
  147400,
  147425,
  147450,
  147475,
  147500,
  147525,
  147550,
  147575,
  147600,   
  144700,   
  144725,
  144750,
  144775,
  144800,
  144825,
  144850,
  144875,
  144900,  
  145075,   
  145100,
  145125,   // 60 = chan 61
  145150,
  145175,
  145200,
  145225,
  145250,
  145275,
  145300,
  145325,
  145350,
  145375,
  145400,   // 72

};

int TXFrequs []      // Edit this for TX frequencies wanted. Added to RX frequs.
{ 0,
  -600,
  -600,
  -600,
  -600,
  -600,
  -600,
  -600,
  -600,
  -600,
  -600,
  -600,
  -600,
  -600,
  -600,
  -600,
  -600,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  600,
  600,
  600,
  600,
  600,
  600,
  600,
  600,
  600,
  600,
  600,
  600,
  600,
  600,
  600,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,      
  0,      
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,    
  0,     
  0,    // 60 = chan 61
  0,
  0,
  0,  
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,      // 72
};

int SubTone []      // Edit this for CTCSS frequencies wanted. NOTE: Converted to int to save space
{ 0,      // "0 ="  chan 1
  0,
  1230,  //  chan 3
  0,
  915,   //  chan 5
  0,
  915,   //  chan 7
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  915,   // chan 15
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  1230,    // chan 26
  0,
  0,
  915,   // chan 29
  915,   // chan 30
  0,
  0,
  0,
  0,
  0,
  915,   // chan 36
  0,
  0,
  0,
  915,   // chan 40
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,     
  0,     
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,    
  0, 
  0,    //"60"= Chan 61
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,        // 72 
};

int sinetable[] // one cycle sine wave via R2R network.
{ 0,1,2,3,5,7,10,13,16,19,22,25,27,28,29,30,31,30,29,28,27,25,22,19,16,13,10,7,5,3,2,1, };

LiquidCrystal_PCF8574 lcd(0x3f); // set the LCD address to 0x27 or 0x3f to suit display interface.

void setup() {
  lcd.setBacklight(255);
  lcd.begin(16, 2);                          // 16 char x 2 line LCD display
  lcd.print(Version);
  delay(1000);
  lcd.setCursor(0, 0);
  lcd.print("RX    Chan=   TX");
  bool i2c_found;

  // Start serial and initialize the Si5351
  i2c_found = clockgen.init(SI5351_CRYSTAL_LOAD_8PF, 0, 0);
  if (!i2c_found)
  {
    lcd.print("Device not found on I2C bus!");
  }

  // configure rotary encoder & push-button
  pinMode(ENCODER_A, INPUT_PULLUP);     // encoder leads
  pinMode(ENCODER_B, INPUT_PULLUP);
  pinMode(REV_SWITCH, INPUT_PULLUP);    // Hi = nomal, LOW =  Reverse
  pinMode(TRX, INPUT_PULLUP);     // Hi = RX, LOW =  TX
  pinMode(Input2, INPUT_PULLUP);
  pinMode(Output1, OUTPUT);
  DDRB  = B111111;                // sets Arduino pins 7 to 12 as outputs.
  PORTB = B100000;                // R2R out at rest to half level.
  pinMode(Output2, OUTPUT);
  digitalWrite(Output1, HIGH);    // output1 off
  digitalWrite(Output2, HIGH);    // output2 off

  encoder = read_encoder();       // returns +1/0/-1 (cw/nc/ccw)
  mode=( digitalRead(REV_SWITCH));
  TRXmode=(digitalRead(TRX));
  set_synth(0,1);
}

/******************************
  main loop
*******************************/
void loop() {
  if ( digitalRead(REV_SWITCH) != mode){
    mode=( digitalRead(REV_SWITCH));
    set_synth(mode, TRXmode);
  }
  if (digitalRead(TRX) !=TRXmode){
    TRXmode=(digitalRead(TRX));
    set_synth(mode, TRXmode);  
  }
  // check the frequency
  encoder = read_encoder();                       // returns +1/0/-1 (cw/nc/ccw)
  if (encoder != 0) {
    channel = channel + encoder;
    channel = constrain(channel, 0, MaxChannel);
    set_synth(mode, TRXmode);
  }
}

/******************************
  read encoder direction
*******************************/
int read_encoder() {
  // definitions
#define DELAY 100                               // number of polling-loops   
  static unsigned short integrator;               // encoder debounce

  /* the array values are determined by combining the previous and current A1,A0
    Gray Code bit-patterns to form an array-index. +1 is assigned to all clockwise
    (CW) indexes; -1 is assigned to all counter-clockwise (CCW) indexes; and 0 is
    assigned to the remaining positions to indicate no movement*/
  static char encoder[] = {0, -1, 1, 0, 1, 0, 0, -1, -1, 0, 0, 1, 0, 1, -1, 0}; // SWAPPED PHASES
  //   static char encoder[] = {0,1,-1,0,-1,0,0,1,1,0,0,-1,0,-1,1,0};
  static unsigned char encoder_index = 0;

  static long count = 0;                         // CW rotation = +1; CCW rotation = -1
  static long this_reading;
  static long last_reading;
  char value;                                    // return value

  /* check for encoder rotation. The absolute "count" value changes as the encoder rotates.
    Typically there are four states (counts) per click, but often more due to contact bounce.
    Bounce doesn't matter so long as the "count" decreases" for CCW rotation; remains constant
    when there is no rotation; and increases with CW rotation. Bounce can be completely
    eliminated if we look for the "detent" code and integrate as shown below. */
  encoder_index <<= 2;                           // shift previous bit-pattern left two places
  encoder_index |= (ENCODER_INPUT_PORT & 0x03);  // get current bit-pattern and combine with previous
  count += encoder[( encoder_index & 0x0f )];

  // debounce encoder
  if ((ENCODER_INPUT_PORT & 0x03) == 3) {        // encoder on an indent (3 decimal == 00000011 binary)
    if (integrator < DELAY) {
      integrator++;
    }
  } else {                                       // encoder not resting on indent
    if (integrator > 0) {
      integrator--;
    }
  }

  // prepare the "return" value
  if (integrator >= DELAY) {                     // encoder deemed to be stationary
    this_reading = count;
    if (this_reading < last_reading) value = -1;
    if (this_reading == last_reading) value = 0;
    if (this_reading > last_reading) value = 1;
    last_reading = this_reading;                 // both readings now hold current "count"
    integrator = 0;                              // reset integrator
  } else {
    value = 0;
  }
  return (value);
}

/*******************************
  set the Si5351 CLK0 frequency
 ******************************/
 
void set_synth( int frmode, int trx )
{  
  lcd.setCursor(10, 0);    // write channel number to the LCD
  if (channel < 9)        // "9" = 10 as array starts at 0
  {
    lcd.print(" ");       // extra space for formalling a single digit.
  }
  lcd.print((channel + 1)); // display the channel number
// Select FWD or REV frequencies
  if (frmode == HIGH)    // test for fwd or rev repeater...
  { RXgenfrequency = (RXFrequs[channel]); // normal operation
    TXgenfrequency = ((RXFrequs[channel]) + (TXFrequs[channel]));
  }
  else
  {
    TXgenfrequency = (RXFrequs[channel]); // reverse repeater operation
    RXgenfrequency = ((RXFrequs[channel]) + (TXFrequs[channel]));
  }
// Then display them
  if ( trx == HIGH) // =RX.
  {
  lcd.setCursor(0,0);
  lcd.print("RX   ");
  lcd.setCursor(14,0);
  lcd.print("  ");
  }  
  rxDisplayFreq = float (RXgenfrequency)/1000;    // display the RX frequency
  lcd.setCursor(0, 1);                   // start of second LCD line
  lcd.print(rxDisplayFreq, 3);
   if ( trx == LOW) // =TX.
  {
  lcd.setCursor(0,0);
  lcd.print("  ");
  lcd.setCursor(14,0);
  lcd.print("TX");
  }
  lcd.setCursor(9, 1);
  txDisplayFreq = float (TXgenfrequency)/1000;
  lcd.print(txDisplayFreq, 3);
  
// calculate the actual frequencies to set

  RXgenfrequency = (((RXgenfrequency * 1000) - IFoffset) / RXmultiplier);
  TXgenfrequency = ((TXgenfrequency  * 1000) / TXmultiplier);

// is it transmittig or receiving?

  if ( trx == HIGH) // =RX.
  { 
    clockgen.output_enable(SI5351_CLK0, 1);                 // turn RX gen on
    clockgen.set_freq((RXgenfrequency * 100), SI5351_CLK0); // set the RX generator frequ,
    clockgen.output_enable(SI5351_CLK1, 0);                 // turn TX gen off
    TIMSK1 = 0;                                             // subtone off
  }
  else                            // =TX.
  { 
    clockgen.output_enable(SI5351_CLK1, 1);                 // turn TX gen on
    clockgen.set_freq((TXgenfrequency * 100), SI5351_CLK1); // set the TX generator frequ,
    clockgen.output_enable(SI5351_CLK0, 0);                 // turn RX gen off
    if ((SubTone[channel]) != 0)
    { SineDivider = (RefFrequ/(float)(SubTone[channel]/10)); // Subtone in as Int = Float/10
    CTCSS_Tone();                                           // subtone on
    lcd.setCursor(0,0);
    lcd.print((float)SubTone[channel]/10,1);                       //subtone as int
    }
    else 
    {
    TIMSK1 = 0;                                             // subtone off
    lcd.setCursor(0,0);
    lcd.print("     ");
    }
  }
}

void CTCSS_Tone()
{
  cli();                    // interupsts off for setup
  TCCR1A = 0;               // 
  TCCR1B = 0; //B00000010;       //  no prescaler
   // set compare match register for 1hz increments
  OCR1A = SineDivider ;// = (16*10^6) / (1*1024) - 1 (must be <65536)
  // turn on CTC mode
  TCCR1B |= (1 << WGM12);
  // Set CS10 no prescaler
  TCCR1B |= (0 << CS12) | (0 << CS11) | (1 << CS10);  
  // enable timer compare interrupt
  TIMSK1 |= (1 << OCIE1A);
  sei();                    // interupsts on      
}
ISR(TIMER1_COMPA_vect){
  PORTB=(sinetable[i++]);   // write value from table to PORTB
  if(i>=32){
    i=0;
  }
}
